'use strict'
const { createCoverageMap } = require('istanbul-lib-coverage')

const { addHook, channel, AsyncResource } = require('./helpers/instrument')
const shimmer = require('../../datadog-shimmer')
const log = require('../../dd-trace/src/log')

const testStartCh = channel('ci:cucumber:test:start')
const testFinishCh = channel('ci:cucumber:test:finish') // used for test steps too

const testStepStartCh = channel('ci:cucumber:test-step:start')

const errorCh = channel('ci:cucumber:error')

const testSuiteStartCh = channel('ci:cucumber:test-suite:start')
const testSuiteFinishCh = channel('ci:cucumber:test-suite:finish')
const testSuiteCodeCoverageCh = channel('ci:cucumber:test-suite:code-coverage')

const itrConfigurationCh = channel('ci:cucumber:itr-configuration')
const skippableSuitesCh = channel('ci:cucumber:test-suite:skippable')
const sessionStartCh = channel('ci:cucumber:session:start')
const sessionFinishCh = channel('ci:cucumber:session:finish')

const {
  getCoveredFilenamesFromCoverage,
  resetCoverage,
  mergeCoverage,
  fromCoverageMapToCoverage,
  getTestSuitePath
} = require('../../dd-trace/src/plugins/util/test')

// We'll preserve the original coverage here
const originalCoverageMap = createCoverageMap()

// TODO: remove in a later major version
const patched = new WeakSet()

let pickleByFile = {}
const pickleResultByFile = {}
let isSuitesSkipped = false

function getSuiteStatusFromTestStatuses (testStatuses) {
  if (testStatuses.some(status => status === 'fail')) {
    return 'fail'
  }
  if (testStatuses.every(status => status === 'skip')) {
    return 'skip'
  }
  return 'pass'
}

function getStatusFromResult (result) {
  if (result.status === 1) {
    return { status: 'pass' }
  }
  if (result.status === 2) {
    return { status: 'skip' }
  }
  if (result.status === 4) {
    return { status: 'skip', skipReason: 'not implemented' }
  }
  return { status: 'fail', errorMessage: result.message }
}

function getStatusFromResultLatest (result) {
  if (result.status === 'PASSED') {
    return { status: 'pass' }
  }
  if (result.status === 'SKIPPED' || result.status === 'PENDING') {
    return { status: 'skip' }
  }
  if (result.status === 'UNDEFINED') {
    return { status: 'skip', skipReason: 'not implemented' }
  }
  return { status: 'fail', errorMessage: result.message }
}

function wrapRun (pl, isLatestVersion) {
  if (patched.has(pl)) return

  patched.add(pl)

  shimmer.wrap(pl.prototype, 'run', run => function () {
    if (!testStartCh.hasSubscribers) {
      return run.apply(this, arguments)
    }

    const asyncResource = new AsyncResource('bound-anonymous-fn')
    return asyncResource.runInAsyncScope(() => {
      const testSuiteFullPath = this.pickle.uri

      if (!pickleResultByFile[testSuiteFullPath]) { // first test in suite
        testSuiteStartCh.publish(testSuiteFullPath)
      }

      const testSourceLine = this.gherkinDocument &&
        this.gherkinDocument.feature &&
        this.gherkinDocument.feature.location &&
        this.gherkinDocument.feature.location.line

      testStartCh.publish({
        testName: this.pickle.name,
        fullTestSuite: testSuiteFullPath,
        testSourceLine
      })
      try {
        const promise = run.apply(this, arguments)
        promise.finally(() => {
          const result = this.getWorstStepResult()
          const { status, skipReason, errorMessage } = isLatestVersion
            ? getStatusFromResultLatest(result) : getStatusFromResult(result)

          if (!pickleResultByFile[testSuiteFullPath]) {
            pickleResultByFile[testSuiteFullPath] = [status]
          } else {
            pickleResultByFile[testSuiteFullPath].push(status)
          }
          testFinishCh.publish({ status, skipReason, errorMessage })
          // last test in suite
          if (pickleResultByFile[testSuiteFullPath].length === pickleByFile[testSuiteFullPath].length) {
            const testSuiteStatus = getSuiteStatusFromTestStatuses(pickleResultByFile[testSuiteFullPath])
            if (global.__coverage__) {
              const coverageFiles = getCoveredFilenamesFromCoverage(global.__coverage__)

              testSuiteCodeCoverageCh.publish({
                coverageFiles,
                suiteFile: testSuiteFullPath
              })
              // We need to reset coverage to get a code coverage per suite
              // Before that, we preserve the original coverage
              mergeCoverage(global.__coverage__, originalCoverageMap)
              resetCoverage(global.__coverage__)
            }

            testSuiteFinishCh.publish(testSuiteStatus)
          }
        })
        return promise
      } catch (err) {
        errorCh.publish(err)
        throw err
      }
    })
  })
  shimmer.wrap(pl.prototype, 'runStep', runStep => function () {
    if (!testStepStartCh.hasSubscribers) {
      return runStep.apply(this, arguments)
    }
    const testStep = arguments[0]
    let resource

    if (isLatestVersion) {
      resource = testStep.text
    } else {
      resource = testStep.isHook ? 'hook' : testStep.pickleStep.text
    }

    const asyncResource = new AsyncResource('bound-anonymous-fn')
    return asyncResource.runInAsyncScope(() => {
      testStepStartCh.publish({ resource })
      try {
        const promise = runStep.apply(this, arguments)

        promise.then((result) => {
          const { status, skipReason, errorMessage } = isLatestVersion
            ? getStatusFromResultLatest(result) : getStatusFromResult(result)

          testFinishCh.publish({ isStep: true, status, skipReason, errorMessage })
        })
        return promise
      } catch (err) {
        errorCh.publish(err)
        throw err
      }
    })
  })
}

function pickleHook (PickleRunner) {
  if (process.env.CUCUMBER_WORKER_ID) {
    // Parallel mode is not supported
    log.warn('Unable to initialize CI Visibility because Cucumber is running in parallel mode.')
    return PickleRunner
  }

  const pl = PickleRunner.default

  wrapRun(pl, false)

  return PickleRunner
}

function testCaseHook (TestCaseRunner) {
  if (process.env.CUCUMBER_WORKER_ID) {
    // Parallel mode is not supported
    log.warn('Unable to initialize CI Visibility because Cucumber is running in parallel mode.')
    return TestCaseRunner
  }

  const pl = TestCaseRunner.default

  wrapRun(pl, true)

  return TestCaseRunner
}

addHook({
  name: '@cucumber/cucumber',
  versions: ['7.0.0 - 7.2.1'],
  file: 'lib/runtime/pickle_runner.js'
}, pickleHook)

addHook({
  name: '@cucumber/cucumber',
  versions: ['>=7.3.0'],
  file: 'lib/runtime/test_case_runner.js'
}, testCaseHook)

function getPicklesToRun (runtime, suitesToSkip) {
  return runtime.pickleIds.filter((pickleId) => {
    const test = runtime.eventDataCollector.getPickle(pickleId)
    return !suitesToSkip.includes(getTestSuitePath(test.uri, process.cwd()))
  }, {})
}

function getPickleByFile (runtime) {
  return runtime.pickleIds.reduce((acc, pickleId) => {
    const test = runtime.eventDataCollector.getPickle(pickleId)
    if (acc[test.uri]) {
      acc[test.uri].push(test)
    } else {
      acc[test.uri] = [test]
    }
    return acc
  }, {})
}

addHook({
  name: '@cucumber/cucumber',
  versions: ['>=7.0.0'],
  file: 'lib/runtime/index.js'
}, (runtimePackage, cucumberVersion) => {
  shimmer.wrap(runtimePackage.default.prototype, 'start', start => async function () {
    const asyncResource = new AsyncResource('bound-anonymous-fn')
    let onDone

    const configPromise = new Promise(resolve => {
      onDone = resolve
    })

    asyncResource.runInAsyncScope(() => {
      itrConfigurationCh.publish({ onDone })
    })

    await configPromise

    const skippableSuitesPromise = new Promise(resolve => {
      onDone = resolve
    })

    asyncResource.runInAsyncScope(() => {
      skippableSuitesCh.publish({ onDone })
    })

    const { err, skippableSuites } = await skippableSuitesPromise

    if (!err) {
      const newPickleIds = getPicklesToRun(this, skippableSuites)
      isSuitesSkipped = newPickleIds.length !== this.pickleIds.length
      this.pickleIds = newPickleIds
    }

    pickleByFile = getPickleByFile(this)

    const processArgv = process.argv.slice(2).join(' ')
    const command = process.env.npm_lifecycle_script || `cucumber-js ${processArgv}`

    asyncResource.runInAsyncScope(() => {
      sessionStartCh.publish({ command, frameworkVersion: cucumberVersion })
    })
    const success = await start.apply(this, arguments)

    let testCodeCoverageLinesTotal

    if (global.__coverage__) {
      try {
        testCodeCoverageLinesTotal = originalCoverageMap.getCoverageSummary().lines.pct
      } catch (e) {
        // ignore errors
      }
      // restore the original coverage
      global.__coverage__ = fromCoverageMapToCoverage(originalCoverageMap)
    }

    asyncResource.runInAsyncScope(() => {
      sessionFinishCh.publish({
        status: success ? 'pass' : 'fail',
        isSuitesSkipped,
        testCodeCoverageLinesTotal
      })
    })
    return success
  })

  return runtimePackage
})
